%{

#include "basic.h"

void yyerror (const char *error);
int yylex ();
extern char *yytext;
extern int sourceLineNumber;

%}

%union
{
  CellProgram *cpa;
  CellState *state;
  CellTransition *transition;
  CellTransitionList *transitionList;
  ChemicalConcentration *concentration;
  ConcentrationList *concentrationList;
  EmissionList *emissionList;
  Emission *emission;
  string *cstring;
  double fnum;
  int inum;
  t_splitKind split;
}

%token <cstring> TOK_IDENTIFIER
%token <inum> TOK_INTEGER
%token <fnum> TOK_FLOAT
%token TOK_EMITS
%token TOK_TO
%token TOK_LTE
%token TOK_LBRACE
%token TOK_RBRACE
%token TOK_NORMAL_X TOK_NORMAL_Y TOK_AXIS TOK_POLARIZED 
%token TOK_STATE TOK_TRANSITIONS TOK_COLOR
%type <cpa> stateDescriptions
%type <state> stateDescription
%type <transition> transition
%type <split> splitKind
%type <concentration> chemicalConcentration;
%type <concentrationList> chemicalConcentrationList;
%type <concentrationList> optChemicalConcentrationList;
%type <transitionList> transitionList;
%type <fnum> number
%type <inum> optColor
%type <emissionList> emissionList optEmissions;
%type <emission> emission;

%start program
%%

program
  : stateDescriptions { g_cellProgram = $1; }
  ;

stateDescriptions
  : stateDescriptions stateDescription 
    { $$ = $1; $$->addState ($2); }
  | 
    { $$ = new CellProgram (); }
  ;

stateDescription
  : TOK_STATE TOK_IDENTIFIER TOK_LBRACE
    optColor
    optEmissions
    TOK_TRANSITIONS transitionList 
    TOK_RBRACE 
    { $$ = new CellState (*$2, $7, $5, $4); }
  ;

optColor
  : TOK_COLOR TOK_INTEGER TOK_INTEGER TOK_INTEGER { $$ = ($4 << 16) + ($3 << 8) + $2; }
  | /* empty */ { $$ = -1; }
  ;

optEmissions
  : TOK_EMITS emissionList 
    { $$ = $2; }
  | /* empty */ { $$ = new EmissionList (); } 
  ;

transitionList
  : transition ';' transitionList 	
    { $3->add ($1); $$ = $3;} 
  |     	     
    { $$ = new CellTransitionList (); }
  ;

transition
  : optChemicalConcentrationList TOK_TO '(' TOK_IDENTIFIER ')'
    { $$ = new CellTransition ($1, $4); }
  | optChemicalConcentrationList TOK_TO '(' TOK_IDENTIFIER ',' TOK_IDENTIFIER ')' splitKind
    { $$ = new CellSplit ($1, $4, $6, $8); }
  ;

splitKind
  : TOK_NORMAL_X  { $$ = SPLIT_NORMAL_1; }
  | TOK_NORMAL_Y  { $$ = SPLIT_NORMAL_2; }
  | TOK_AXIS      { $$ = SPLIT_AXIS; } 
  | TOK_POLARIZED { $$ = SPLIT_POLARIZED; }
  ;

number
  : TOK_FLOAT { $$ = $1; }
  | TOK_INTEGER { $$ = (float) $1; }
  ;
	
emissionList
  : emission ';' emissionList
    { $3->add ($1); $$ = $3; }
  | emission
    { $$ = new EmissionList (); $$->add ($1); }
  ;

emission
  : '(' TOK_IDENTIFIER ',' number ')'
    { $$ = new Emission (*$2, $4); }
  ;

optChemicalConcentrationList
  : chemicalConcentrationList { $$ = $1; }
  | { $$ = NULL; }
  ;

chemicalConcentrationList
  : chemicalConcentration '&' chemicalConcentrationList 
    { $3->add ($1); $$ = $3; }
  | chemicalConcentration 
    { $$ = new ConcentrationList ();
      $$->add ($1); 
    }
  ;

chemicalConcentration
 : '(' number TOK_LTE TOK_IDENTIFIER TOK_LTE number ')'
   { $$ = new ChemicalConcentration (*$4, $2, $6); }
  ;

%%

void yyerror (const char *s)
{
  static int numErrors=0;
  if(numErrors<10)
     fprintf (stderr, "Parse error: %s at line %d : Parsing stopped at string %s\n", s, sourceLineNumber, yytext);
  else
  {
     fprintf (stderr, "Quitting after %d errors\n", numErrors);
     exit (1);
  }
}





