/*
** SimInstruction.cc
*/

# include "basic.h"

SimInstruction *SimInstruction::fromString (char *s)
{
  char *spc = strchr (s, ':');

  if (spc == NULL) {
    cerr << "Bad instruction: " << s << endl;
    exit (-1);
  }

  *spc = '\0';
  char *step = new char [(size_t) (spc - s) + 1];
  int stepno;
  strcpy (step, s);

  if (sscanf (step, "%d", &stepno) != 1) {
    cerr << "Bad instruction: " << s << endl;
    exit (-1);
  }

  // cerr << "Step: " << stepno << endl;
  spc++;
  while (*spc == ' ' || *spc == '\t') {
    spc++;
  }

  // cerr << "Rest: " << spc << endl;
  char *scmd = spc;
  spc = strchr (spc, ' ');
  char *cmd;

  if (spc == NULL) {
      cmd = scmd;
  } else {
      *spc = '\0';
      spc++;
      
      cmd = new char [(size_t) (spc - scmd) + 1];
      strcpy (cmd, scmd);
  }
  
  // cerr << "Creating sim instruction: " << cmd << endl;
  
  SimInstruction *res;

  if (strncmp (cmd, "new", 3) == 0) {
      res = new NewCellInstruction (spc);
  } else if (strncmp (cmd, "random", 5) == 0) {
      res = new RandomCellsInstruction (spc);
  } else if (strncmp (cmd, "kill", 4) == 0) {
      res = new KillInstruction (spc);
  } else if (strncmp (cmd, "store", 5) == 0) {
      res = new StoreInstruction ();
  } else if (strncmp (cmd, "stop", 4) == 0) {
      res = new StopInstruction ();
  } else {
      cerr << "Bad instruction: " << cmd << endl;
      exit (-1);
  }
  
  res->step = stepno;
  return res;
}

string SimInstruction::toString ()
{
  return "<sim instruction>";
}

void SimInstruction::interpret (CellSimulator *sim)
{
  // nothing to do
  ;
}

NewCellInstruction::NewCellInstruction (char *s)
{
  if (sscanf (s, "%ld %f %f %f %f %f %f",
              &state, &x, &y, &z, &r, &axis, &normal) != 7)
    {
      cerr << "Bad instruction: new " << s << endl;
    }
}

NewCellInstruction::NewCellInstruction (CellStateId p_state,
                                        float p_x, float p_y, float p_z, float p_r,
                                        float p_axis, float p_normal)
{
  state = p_state;
  x = p_x;
  y = p_y;
  z = p_z;
  r = p_r;
  axis = p_axis;
  normal = p_normal;
}

void NewCellInstruction::interpret (CellSimulator *sim)
{
  SimCell *cell = sim->getNewCell ();
  cell->cell_state = state;
  cell->x = x;
  cell->y = y;
  cell->z = z;
  cell->r = r;
  cell->abAxis = axis;
  cell->abNormal = normal;
}

RandomCellsInstruction::RandomCellsInstruction (char *s)
{
  if (sscanf (s, "%d %ld %ld %f %f %f %f %f %f %f %f %f %f %f %f",
              &ncells,
              &minstate, &maxstate,
              &minx, &maxx,
              &miny, &maxy,
              &minz, &maxz,
              &minr, &maxr,
              &minaxis, &maxaxis,
              &minnormal, &maxnormal) != 15)
    {
      cerr << "Bad instruction: random " << s << endl;
    }
}

static double randomInRange (double low, double high) 
{
  double range = high - low;

  if (range == 0) return low;

  int r = rand ();
  double rfrac = (double) r / (double) RAND_MAX;

  return low + (range * rfrac);
}

static int randomInRangeInt (int low, int high) 
{
  int range = high - low;
  if (range == 0) return low;
  int r = rand ();
  return low + (r % range);
}

void RandomCellsInstruction::interpret (CellSimulator *sim)
{
  for (int i = 0; i < ncells; i++) {
    SimCell *cell = sim->getNewCell ();
    cell->cell_state = randomInRangeInt ((int) minstate, (int) maxstate);
    cell->x = randomInRange (minx, maxx);
    cell->y = randomInRange (miny, maxy);
    cell->z = randomInRange (minz, maxz);
    cell->r = randomInRange (minr, maxr);
    cell->abAxis = randomInRange (minaxis, maxaxis);
    cell->abNormal = randomInRange (minnormal, maxnormal);
  }
}

void KillInstruction::interpret (CellSimulator *sim)
{
  int numkilled = 0;

  for (int i = 0; i < sim->numCells (); i++) {
    SimCell *cell = sim->getCell (i);
    
    if (cell->isAlive () && (cell->getDistanceFrom (x, y, z) <= r)) {
      cell->kill ();
      numkilled++;
    }
  }

  cerr << "Killed: " << numkilled << endl;
}

KillInstruction::KillInstruction (char *s)
{
  if (sscanf (s, "%f %f %f %f", &x, &y, &z, &r) != 4)
    {
      cerr << "Bad instruction: kill " << s << endl;
    }
}

KillInstruction::KillInstruction (float p_x, float p_y, float p_z, float p_r)
{
  x = p_x;
  y = p_y;
  z = p_z;
  r = p_r;
}

StoreInstruction::StoreInstruction () : SimInstruction ()
{
  ;
}

void StoreInstruction::interpret (CellSimulator *sim)
{
  ;
}

StopInstruction::StopInstruction () : SimInstruction ()
{
  ;
}

extern void exitProgram(); // in cellsim.cc

void StopInstruction::interpret (CellSimulator *sim)
{
  exitProgram ();
}






