/*
** SimInstructions.cc
*/

#include "basic.h"
#include "../threads/thread.h"
#include "display.h"

SimInstructions::SimInstructions ()
{
  nelements = 0;
  nspace = BASESIZE; 
  elements = new SimInstruction *[nspace];
  current = 0;
}

SimInstructions::~SimInstructions ()
{
  int i;

  for (i = 0; i < nelements; i++) {
    delete elements[i];
  }

  delete elements;
}

void SimInstructions::grow ()
{
  int i;
  SimInstruction **newelements; 
  size_t newspace = nelements + BASESIZE;
  newelements = new SimInstruction *[newspace];

  assert (newelements != NULL);

  nspace = newspace - nelements;

  for (i = 0; i < nelements; i++)
    {
      newelements[i] = elements[i];
    }
  
  delete elements;
  elements = newelements;
}

void SimInstructions::add (SimInstruction *inst)
{
  if (nspace == 0) grow ();

  assert (nspace > 0);
  
  if (nelements > 0) {
    if (inst->getStep () < elements[nelements - 1]->getStep ()) {
      cerr << "Instruction out of order: " << inst << endl;
      exit (-1);
    }
  }

  nspace--;
  elements[nelements] = inst;
  nelements++;
}

void SimInstructions::loadInstructions (FILE *f)
{
  // each line in the sim instructions file is an instruction:
  //
  //      <step>: <action> [ parameters ]
  //
  // or a comment (beginning with #)
  //
  // The steps must be in order.
  //
  // Actions:
  //    new state x y z r axis normal
  //        Place a new cell at (x, y, z) with radius r, and orientation axis, normal
  //        State is the number of the state (not its name for now)
  //    random n minstate maxstate minx maxx miny maxy minz maxz 
  //             minr maxr minaxis maxaxis minnormal maxnormal
  //        Place n new cells with random properties in the ranges given.
  //        Random ranges are [min, max).
  //    kill x y z r
  //        Kill all cells within distance r of x, y, z
  //    store 
  //        Save the current state in out.<step>
  //    stop 
  //        Stop the simulation and terminate the program
  //    

  char *s = new char [MAX_LINE];
  char *res;

  while ((res = fgets (s, MAX_LINE, f)) != NULL)
    {
      if (*res == '#') {
        ; // Skip comment line
      } else {
        if (strlen (res) == MAX_LINE - 1) {
          cerr << "Fatal error: line too long: " << res;
          exit (-1);
        }
        
        SimInstruction *inst = SimInstruction::fromString (res);
        add (inst);
      }
    }
}

void SimInstructions::interpretInstructions (CellSimulator *sim)
{
  while (current < nelements) 
    {
      SimInstruction *cinst = elements [current];

		if (cinst->getStep ()  < g_tick) {
        cerr << "Encountered an out-of-order instruction\n";
        exit (3);
		}
      
		if (cinst->getStep () ==  g_tick) {
        current++;
        cinst->interpret (sim);
      } else {
        break;
      }
    }
}
